from typing import Callable, Union

import torch

from ncut_pytorch.ncuts.ncut_nystrom import ncut_fn, nystrom_propagate
from ncut_pytorch.utils.math import rbf_affinity


class Ncut:
    """
    Class interface for Normalized Cut, save states of nystrom approximation, can be used to transform new data.
    """
    
    def __init__(
            self,
            n_eig: int = 100,
            track_grad: bool = False,
            d_gamma: float = None,
            device: str = None,
            affinity_fn: Callable[[torch.Tensor, torch.Tensor, float], torch.Tensor] = rbf_affinity,
            **kwargs,
    ):
        """
        
        Args:       
            n_eig (int): number of eigenvectors
            track_grad (bool): keep track of pytorch gradients
            d_gamma (float): affinity gamma parameter, lower d_gamma results in a sharper eigenvectors
            device (str): device, default 'auto'

        Examples:
            >>> from ncut_pytorch import Ncut
            >>> import torch
            >>> X = torch.rand(10000, 100)
            >>> 
            >>> # Method 1: Function-like call
            >>> eigvec = Ncut(X, n_eig=20)
            >>> print(eigvec.shape)  # (10000, 20)
            >>> 
            >>> # Method 2: Create instance and call
            >>> ncut = Ncut(n_eig=20)
            >>> eigvec = ncut(X)  # returns just eigvec
            >>> eigval = ncut.eigval  # access eigenvalues separately
            >>> print(eigvec.shape, eigval.shape)  # (10000, 20) (20,)
            >>> 
            >>> # Method 2.1: Transform new data after fitting
            >>> new_X = torch.rand(500, 100)
            >>> new_eigvec = ncut.transform(new_X)
            >>> print(new_eigvec.shape)  # (500, 20)
        """
        self.n_eig = n_eig
        self.d_gamma = d_gamma
        self.device = device
        self.track_grad = track_grad
        self.affinity_fn = affinity_fn
        self.kwargs = kwargs

        self._nystrom_x = None
        self._nystrom_eigvec = None
        self._eigval = None

    @property
    def eigval(self) -> torch.Tensor:
        return self._eigval

    def fit(self, X: torch.Tensor) -> "Ncut":
        """
        Fit the Ncut model to the input features. save states of nystrom approximation.
        
        Args:
            X (torch.Tensor): input features, shape (N, D)
        Returns:
            ncut (Ncut): Ncut instance
        """
        eigvec, eigval, indices, gamma = \
            ncut_fn(
                X,
                n_eig=self.n_eig,
                d_gamma=self.d_gamma,
                device=self.device,
                track_grad=self.track_grad,
                no_propagation=True,
                affinity_fn=self.affinity_fn,
                **self.kwargs
            )
        # store Ncut state to use in transform()
        self._nystrom_x = X[indices]
        self._nystrom_eigvec = eigvec
        self._eigval = eigval
        self.gamma = gamma
        return self

    def transform(self, X: torch.Tensor) -> torch.Tensor:
        """
        Transform new data using the fitted Ncut model and it's saved states of nystrom approximation.
        
        Args:
            X (torch.Tensor): input features, shape (N, D)
            
        Returns:
            eigvec (torch.Tensor): eigenvectors, shape (N, n_eig)
        """
        # check if fit has been called
        if self._nystrom_x is None:
            raise ValueError("Ncut has not been fitted yet. Call fit() first.")

        # propagate eigenvectors from subgraph to full graph
        eigvec = nystrom_propagate(
            self._nystrom_eigvec,
            X,
            self._nystrom_x,
            gamma=self.gamma,
            device=self.device,
            track_grad=self.track_grad,
            affinity_fn=self.affinity_fn,
            **self.kwargs
        )
        return eigvec

    def fit_transform(self, X: torch.Tensor) -> torch.Tensor:
        """
        
        Args:
            X (torch.Tensor): input features, shape (N, D)

        Returns:
            eigvec (torch.Tensor): eigenvectors, shape (N, n_eig)
        """
        return self.fit(X).transform(X)

    def __new__(cls, X: torch.Tensor = None, n_eig: int = 100, track_grad: bool = False, d_gamma: float = None,
                device: str = None, affinity_fn: Callable[[torch.Tensor, torch.Tensor, float], torch.Tensor] = rbf_affinity, 
                **kwargs) -> Union["Ncut", torch.Tensor]:
        if X is not None:
            # function-like behavior
            eigvec, eigval = ncut_fn(X, n_eig=n_eig, track_grad=track_grad, d_gamma=d_gamma, device=device, affinity_fn=affinity_fn, **kwargs)
            return eigvec
        # normal class instantiation
        return super().__new__(cls)

    def __call__(self, X: torch.Tensor) -> torch.Tensor:
        return self.fit_transform(X)



# Credits: https://github.com/tldr-group/HR-Dv2

from functools import partial
from itertools import product
from typing import List, Literal, Tuple, Union, Optional

import numpy as np
import torch
import torchvision.transforms as transforms
from PIL import Image

PartialTrs = List[partial]
ShiftPattern = Literal["Neumann", "Moore", "Cross"]

# ==================== MODEL TRANSFORMS ====================
def compute_shift_directions(
    pattern: ShiftPattern
) -> List[Tuple[int, int]]:
    # Precompute neighbourhood shift unit vectors
    shifts = [  # shifts in yx format
        (-1, -1),
        (-1, 0),
        (-1, 1),
        (0, -1),
        (0, 0),
        (0, 1),
        (1, -1),
        (1, 0),
        (1, 1),
    ]
    shift_directions: List[Tuple[int, int]] = []
    for i in range(9):
        if pattern == "Neumann" and i % 2 == 1:
            shift_directions.append(shifts[i])
        elif pattern == "Moore" and i != 4:
            shift_directions.append(shifts[i])
        elif pattern == "Cross" and i % 2 == 0:
            shift_directions.append(shifts[i])
    return shift_directions


def get_shift_transforms(
    dists: List[int], 
    patterns: Union[List[ShiftPattern], ShiftPattern, None] = None
) -> Tuple[PartialTrs, PartialTrs]:
    transforms: PartialTrs = [true_iden_partial]
    inv_transforms: PartialTrs = [true_iden_partial]

    def roll_arg_rev(shift: Tuple[int, int], x: torch.Tensor) -> torch.Tensor:
        return torch.roll(x, shift, dims=(-2, -1))
    
    if patterns is None:
        # rotate between the 3 patterns
        patterns = ["Moore", "Neumann", "Cross"] * (len(dists) // 3 + 1)
        patterns = patterns[:len(dists)]

    if isinstance(patterns, str):
        patterns = [patterns] * len(dists)
    

    for d, pattern in zip(dists, patterns):
        shifts = compute_shift_directions(pattern)
        for s in shifts:
            shift = (d * s[0], d * s[1])
            inv_shift = (-d * s[0], -d * s[1])
            tr = partial(roll_arg_rev, shift)
            inv_tr = partial(roll_arg_rev, inv_shift)
            transforms.append(tr)
            inv_transforms.append(inv_tr)
    return transforms, inv_transforms


def iden(x: torch.Tensor) -> torch.Tensor:
    if len(x.shape) == 3:  # batch
        x = x.unsqueeze(0)
    return x


def true_iden(x: torch.Tensor) -> torch.Tensor:
    return x


iden_partial = partial(iden)
true_iden_partial = partial(true_iden)


def get_flip_transforms() -> Tuple[PartialTrs, PartialTrs]:
    def flip_arg_rev(dims: Tuple[int, ...], x: torch.Tensor) -> torch.Tensor:
        if len(x.shape) == 3:  # batch
            x = x.unsqueeze(0)
        return torch.flip(x, dims)

    # np.flip would be faster apparently

    horizontal_flip_partial = partial(flip_arg_rev, (-1,))
    vertical_flip_partial = partial(flip_arg_rev, (-2,))
    transforms = [
        iden_partial,
        horizontal_flip_partial,
        # vertical_flip_partial,
    ]
    inv_tranforms = [
        iden_partial,
        horizontal_flip_partial,
        # vertical_flip_partial,
    ]
    return transforms, inv_tranforms


def get_rotation_transforms() -> Tuple[PartialTrs, PartialTrs]:
    def rot_arg_rev(dims: Tuple[int, ...], k: int, x: torch.Tensor) -> torch.Tensor:
        if len(x.shape) == 3:  # batch
            x = x.unsqueeze(0)
        return torch.rot90(x, k, dims)

    transforms = [iden_partial]
    inv_transforms = [iden_partial]
    for direction in [-1, 1]:
        for angle in [1, 2, 3]:
            fwd = partial(rot_arg_rev, (-2, -1), direction * angle)
            inv = partial(rot_arg_rev, (-2, -1), -direction * angle)
            if direction == 1 and angle == 2:
                pass
            else:
                transforms.append(fwd)
                inv_transforms.append(inv)
    return transforms, inv_transforms


def combine_transforms(
    fwd_1: PartialTrs, fwd_2: PartialTrs, inv_1: PartialTrs, inv_2: PartialTrs
) -> Tuple[PartialTrs, PartialTrs]:
    def combined(tr_1, tr_2, x: torch.Tensor) -> torch.Tensor:
        return tr_2(tr_1(x))

    fwd_pairs = product(fwd_1, fwd_2)
    inv_pairs = product(inv_1, inv_2)

    fwd: PartialTrs = []
    inv: PartialTrs = []
    for pair in fwd_pairs:
        f1, f2 = pair
        partial_combined = partial(combined, f1, f2)
        fwd.append(partial_combined)
    for pair in inv_pairs:
        i1, i2 = pair
        partial_combined = partial(combined, i1, i2)
        inv.append(partial_combined)
    return (fwd, inv)


def combine_transforms_pairwise(
    fwd_1: PartialTrs, fwd_2: PartialTrs, inv_1: PartialTrs, inv_2: PartialTrs
) -> Tuple[PartialTrs, PartialTrs]:
    """Combines transforms in a pairwise fashion (Aa, Bb, Ca, Db) rather than taking the product.
    If lengths are unequal, the shorter list's last element is repeated to match the longer list.
    """
    def combined(tr_1, tr_2, x: torch.Tensor) -> torch.Tensor:
        return tr_2(tr_1(x))

    fwd: PartialTrs = []
    inv: PartialTrs = []
    
    i_fwd_2 = 0
    i_inv_2 = 0
    for i in range(len(fwd_1)):
        fwd.append(partial(combined, fwd_1[i], fwd_2[i_fwd_2]))
        inv.append(partial(combined, inv_1[i], inv_2[i_inv_2]))
        i_fwd_2 = (i_fwd_2 + 1) % len(fwd_2)
        i_inv_2 = (i_inv_2 + 1) % len(inv_2)

    return (fwd, inv)


# ==================== PYTORCH INPUT TRANSFORMS ====================
def get_input_transform(resize: Union[int, Tuple[int, int]] = 512, crop: Optional[int] = None) -> transforms.Compose:
    if crop is None:
        crop = resize
    
    transform = transforms.Compose(
        [
            transforms.Resize(resize),
            transforms.CenterCrop(crop),
            transforms.ToTensor(),
            transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)),
        ]
    )
    return transform


to_norm_tensor = transforms.Compose(
    [
        transforms.ToTensor(),
        transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)),
    ]
)

unnormalize = transforms.Normalize(
    mean=(-0.485 / 0.229, -0.456 / 0.224, -0.406 / 0.225),
    std=(1 / 0.229, 1 / 0.224, 1 / 0.225),
)


def centre_crop(crop_h: int, crop_w: int) -> transforms.Compose:
    transform = transforms.Compose(
        [
            transforms.CenterCrop((crop_h, crop_w)),
            transforms.ToTensor(),
            transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)),
        ]
    )
    return transform


def closest_crop(
    h: int, w: int, patch_size: int = 14, to_tensor: bool = True
) -> transforms.Compose:
    # Crop to h,w values that are closest to given patch/stride size
    sub_h: int = h % patch_size
    sub_w: int = w % patch_size
    new_h, new_w = h - sub_h, w - sub_w
    if to_tensor:
        transform = transforms.Compose(
            [
                transforms.CenterCrop((new_h, new_w)),
                transforms.ToTensor(),
                transforms.Normalize(
                    mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)
                ),
            ]
        )
    else:
        transform = transforms.Compose(
            [
                transforms.CenterCrop((new_h, new_w)),
            ]
        )
    return transform


def closest_pad(h: int, w: int, patch_size: int = 14) -> transforms.Compose:
    add_h: int = patch_size - (h % patch_size)
    add_w: int = patch_size - (w % patch_size)
    transform = transforms.Compose(
        [
            transforms.Pad((add_w, add_h, 0, 0)),
        ]
    )
    return transform


to_img = transforms.ToPILImage()
to_tensor = transforms.ToTensor()


def load_image(
    path: str, transform: transforms.Compose
) -> Tuple[torch.Tensor, Image.Image]:
    # Load image with PIL, convert to tensor by applying $transform, and invert transform to get display image
    image = Image.open(path).convert("RGB")
    tensor = transform(image)
    transformed_img = to_img(unnormalize(tensor))
    return tensor, transformed_img


def to_numpy(x: torch.Tensor, batched: bool = True) -> np.ndarray:
    if batched:
        x = x.squeeze(0)
    return x.detach().cpu().numpy()


def flatten(
    x: np.ndarray, h: int, w: int, c: int, convert: bool = False
) -> np.ndarray:
    if type(x) == torch.Tensor:
        x = to_numpy(x)
    x = x.reshape((c, h * w))
    x = x.T
    return x  # type: ignore



from typing import List, Tuple, Optional, Union

import cv2
import numpy as np
import torch
import torch.nn as nn
from PIL import Image
from torchvision import transforms

from .predictor import NcutPredictor, NotInitializedError


class NcutVisionPredictor:
    _initialized: bool = False

    def __init__(self,
                 model: nn.Module,
                 transform: transforms.Compose,
                 batch_size: int):
        self.model = model
        self.transform = transform

        self.batch_size = batch_size

        self._images: List[Image.Image]
        self._image_whs: List[Tuple[int, int]]
        self._feat_hws: Tuple[int, int]

        self.predictor = NcutPredictor()

    def set_images(self,
                   images: List[Image.Image],
                   n_segments: List[int] = (5, 25, 50, 100, 250)):
        """
        set the images and save its features in the cache.
        
        Args:
            images (List[Image.Image]): List of images to set.
            n_segments (List[int], optional): Number of segments to cache.
                n_segments is showed in the preview function.
        """
        features = self.forward_model(images)  # (b, c, h, w)
        self._images = images
        self._image_whs = np.array([image.size for image in images])
        self._feat_hws = (features.shape[2], features.shape[3])

        flat_features = features.permute(0, 2, 3, 1).reshape(-1, features.shape[1])
        self.predictor.initialize(flat_features, n_segments)
        self._initialized = True

    @torch.inference_mode()
    def forward_model(self, images: List[Image.Image]) -> torch.Tensor:
        device = next(self.model.parameters()).device
        with torch.autocast(device_type=device.type, enabled=True):
            all_features = []
            for i in range(0, len(images), self.batch_size):
                batch_images = images[i:i + self.batch_size]
                transformed_images = torch.stack([self.transform(image) for image in batch_images])
                transformed_images = transformed_images.to(device)
                features = self.model(transformed_images)
                features = features.to('cpu')
                all_features.append(features)
        return torch.cat(all_features, dim=0)

    def generate(self, n_segment: int) -> torch.Tensor:
        """
        generate the cluster assignment for the images.
        
        Args:
            n_cluster (int): Number of clusters to generate.
            
        Returns:
            torch.Tensor: Cluster assignment for the images. (b, h, w)
        """
        self.__check_initialized()
        cluster_assignment = self.predictor.get_n_segments(n_segment)
        b, h, w = len(self._images), self._feat_hws[0], self._feat_hws[1]
        cluster_assignment = cluster_assignment.reshape(b, h, w)
        return cluster_assignment

    def preview(self,
                point_coord: Tuple[int, int],
                image_indices: int) -> List[torch.Tensor]:
        """
        preview the hierarchy cluster assignment for the images.
        
        Args:
            point_coord (Tuple[int, int]): The coordinate of the point to preview.
            image_indices (int): The index of the image to preview.
            
        Returns:
            List[torch.Tensor]: List of masks for each hierarchy level. each mask is (b, h, w)
        """
        self.__check_initialized()
        b, h, w = len(self._images), self._feat_hws[0], self._feat_hws[1]

        point_index = self._image_xy_to_tensor_index(self._image_whs, 
                                                     self._feat_hws,
                                                     np.array([point_coord]), 
                                                     np.array([image_indices])
                                                     )[0]
        masks = self.predictor.get_hierarchy_masks(point_index)
        masks = [mask.reshape(b, h, w) for mask in masks]
        return masks

    def summary(self,
                n_segments: List[int] = (5, 25, 50, 100, 250),
                draw_border: bool = True,
                ) -> List[torch.Tensor]:
        """
        summary the cluster assignment for the images.
        
        Args:
            n_segments (List[int]): Number of segments to summary.
        """
        self.__check_initialized()
        display_hw = 512
        
        colors = []
        colors.append(self._images)
        for n_segment in n_segments:
            cluster_assignment = self.generate(n_segment)
            color = self.color_discrete(cluster_assignment, draw_border=draw_border)
            colors.append(color)
        color = self.color_continues()
        colors.append(color)
        
        # make a grid of images
        n_rows = len(self._images)
        n_cols = len(n_segments) + 2
        grid_image = Image.new('RGB', size=(n_cols * display_hw, n_rows * display_hw))
        for i in range(n_rows):
            for j in range(n_cols):
                img = colors[j][i]
                img = Image.fromarray(np.array(img))
                img = img.resize((display_hw, display_hw), Image.Resampling.NEAREST)
                grid_image.paste(img, box=(j * display_hw, i * display_hw))
        return grid_image

    def predict(self,
                point_coords: np.ndarray,
                point_labels: np.ndarray,
                image_indices: np.ndarray,
                click_weight: float = 0.5,
                **kwargs) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        predict the mask and heatmap for the images, based on the clicks.
        
        Args:
            point_coords (np.ndarray): The coordinates of the points to predict. (n, 2)
            point_labels (np.ndarray): The labels of the points to predict. (n, )
            image_indices (np.ndarray): The indices of the images to predict. (n, )
            click_weight (float, optional): The weight of the click. Defaults to 0.5.
            
        Returns:
            Tuple[torch.Tensor, torch.Tensor]: Mask and heatmap for the images. (b, h, w)
        """
        self.__check_initialized()
        fg_indices = self._image_xy_to_tensor_index(self._image_whs, self._feat_hws,
                                                    point_coords[point_labels == 1],
                                                    image_indices[point_labels == 1])
        bg_indices = self._image_xy_to_tensor_index(self._image_whs, self._feat_hws,
                                                    point_coords[point_labels == 0],
                                                    image_indices[point_labels == 0])
        b, h, w = len(self._images), self._feat_hws[0], self._feat_hws[1]

        mask, heatmap = self.predictor.predict_clicks(fg_indices, bg_indices, click_weight, **kwargs)

        mask = mask.reshape(b, h, w)
        heatmap = heatmap.reshape(b, h, w)

        return mask, heatmap

    def inference(self, images: List[Image.Image]) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        inference the mask and heatmap for new images, based on the saved states in the predict function.
        
        Args:
            images (List[Image.Image]): List of images to inference.
            
        Returns:
            Tuple[torch.Tensor, torch.Tensor]: Mask and heatmap for the images. (b, h, w)
        """
        self.__check_initialized()
        new_features = self.forward_model(images)
        b, c, h, w = new_features.shape
        new_features = new_features.permute(0, 2, 3, 1).reshape(-1, c)

        mask, heatmap = self.predictor.inference_new_features(new_features)

        mask = mask.reshape(b, h, w)
        heatmap = heatmap.reshape(b, h, w)
        return mask, heatmap

    def color_continues(self) -> np.ndarray:
        """
        color the features by continues mspace color palette.
        
        Returns:
            np.ndarray: RGB image. (b, h, w, 3)
        """
        self.__check_initialized()
        b, h, w = len(self._images), self._feat_hws[0], self._feat_hws[1]
        color_palette = self.predictor.get_color_palette()
        rgb = color_palette.reshape(b, h, w, 3)
        rgb = (rgb * 255).to(torch.uint8).cpu().numpy()
        return rgb

    def color_discrete(self, 
                       cluster_assignment: torch.Tensor,
                       draw_border: bool = True,
                       ) -> List[Image.Image]:
        """
        color the features by discrete mspace color palette.
        
        Args:
            cluster_assignment (torch.Tensor): Cluster assignment for the images. (b, h, w)
            draw_boundaries (bool, optional): Whether to draw boundaries. Defaults to True.
            
        Returns:
            List[Image.Image]: List of RGB images.
        """
        self.__check_initialized()
        b, h, w = cluster_assignment.shape
        color_palette = self.predictor.get_color_palette()
        n_cluster = cluster_assignment.max() + 1
        cluster_assignment = cluster_assignment.flatten()
        discrete_rgb = np.zeros((b * h * w, 3), dtype=np.uint8)
        for i in range(n_cluster):
            mask = cluster_assignment == i
            color = color_palette[mask].mean(0)
            color = (color * 255).cpu().numpy()
            color = np.clip(color, 0, 255).astype(np.uint8)
            discrete_rgb[mask] = color
        discrete_rgb = discrete_rgb.reshape(b, h, w, 3)
        
        # convert to PIL image and resize to the original size
        pil_images = []
        for i in range(b):
            img = Image.fromarray(discrete_rgb[i])
            img = img.resize(self._images[i].size, Image.Resampling.NEAREST)
            if draw_border:
                img = self._draw_segments_border(img)
            pil_images.append(img)
        
        return pil_images
    
    def refresh_color_palette(self):
        self.predictor.refresh_color_palette()
    
    @staticmethod
    def _draw_segments_border(img: Image.Image, min_area_ratio: float = 0.0005) -> Image.Image:
        img = np.array(img)
        src_gray = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
        src_gray = src_gray.astype(np.int32)
    
        contours, hierarchy = cv2.findContours(src_gray, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
    
        area_threshold = src_gray.shape[0] * src_gray.shape[1] * min_area_ratio
        drawing = img.copy()
        for i in range(len(contours)):
            area = cv2.contourArea(contours[i])
            if area >= area_threshold:
                color = (0, 0, 0)
                cv2.drawContours(drawing, contours, i, color, 1, cv2.LINE_8, hierarchy, 0)
        return Image.fromarray(drawing)

    @staticmethod
    def _image_xy_to_tensor_index(image_whs: np.array,
                                  feat_hws: np.array,
                                  point_coords: np.ndarray,
                                  image_indices: np.ndarray) -> np.ndarray:
        """
        Convert image xy coordinates to tensor index.
        Args:
            image_whs: List of image width and height, (n_images, 2)
            feat_hws: Feature width and height, (2, )
            point_coords: Point coordinates, (n_points, 2)
            image_indices: Image indices for each point, (n_points, )
        Returns:
            Point indices
        """
        if len(point_coords) == 0:
            return np.array([], dtype=np.int64)

        wh = image_whs[image_indices]
        point_coords = point_coords / wh

        point_coords = np.flip(point_coords, axis=1)  # (x, y) -> (y, x)

        point_coords = point_coords * feat_hws
        point_coords = point_coords.astype(np.int64)

        point_indices = point_coords[:, 0] * feat_hws[0] + point_coords[:, 1]

        offset_perimg = np.prod(feat_hws)
        offsets = image_indices * offset_perimg
        point_indices = point_indices + offsets
        point_indices = point_indices.astype(np.int64)
        return point_indices

    def __check_initialized(self):
        if not self._initialized:
            raise NotInitializedError("Not initialized, please call set_images() first")

    def to(self, device: Union[str, torch.device]):
        self.model = self.model.to(device)
        self.predictor = self.predictor.to(device)
        return self


__all__ = ['ncut_fn', 'nystrom_propagate']

from typing import Callable, Union

import torch

from ncut_pytorch.utils.gamma import find_gamma_by_degree_after_fps
from ncut_pytorch.utils.math import rbf_affinity, gram_schmidt, normalize_affinity, svd_lowrank, correct_rotation, \
    keep_topk_per_row
from ncut_pytorch.utils.sample import farthest_point_sampling
from ncut_pytorch.utils.device import auto_device


class NystromConfig:
    """
    Internal configuration for nystrom approximation, can be overridden by kwargs
    Values are optimized based on empirical experiments, no need to change the values
    """
    n_sample = 10240  # number of samples for nystrom approximation, 10240 is large enough for most cases
    n_sample_max_ratio = 1/4  # max ratio of n_sample to n_data, not full sample ensures balanced sampling
    n_sample2 = 1024  # number of samples for eigenvector propagation, 1024 is large enough for most cases
    n_neighbors = 32  # number of neighbors for eigenvector propagation, 10 is large enough for most cases
    n_neighbors_max_ratio = 1/32  # max ratio of n_neighbors to n_sample2, to avoid over smoothing
    matmul_chunk_size = 65536  # chunk size for matrix multiplication, larger chunk size is faster but requires more memory
    move_output_to_cpu = True  # if True, will move output to cpu, saves VRAM
    
    def update(self, kwargs: dict):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise ValueError(f"Invalid kwarg: {key}")


def ncut_fn(
        X: torch.Tensor,
        n_eig: int = 100,
        track_grad: bool = False,
        d_gamma: float = 'auto',
        device: str = None,
        gamma: float = None,
        make_orthogonal: bool = False,
        affinity_fn: Callable[[torch.Tensor, torch.Tensor, float], torch.Tensor] = rbf_affinity,
        no_propagation: bool = False,
        **kwargs,
) -> Union[tuple[torch.Tensor, torch.Tensor], tuple[torch.Tensor, torch.Tensor, torch.Tensor, float]]:
    """Normalized Cut, balanced sampling and nystrom approximation.

    Args:
        X (torch.Tensor): input features, shape (N, D)
        n_eig (int): number of eigenvectors
        track_grad (bool): keep track of pytorch gradients
        d_gamma (float): affinity gamma parameter, lower d_gamma results in sharper eigenvectors
        device (str): device, default 'auto' (auto detect GPU)
        gamma (float): affinity parameter, override d_gamma if provided
        make_orthogonal (bool): make eigenvectors orthogonal
        affinity_fn (callable): affinity function, default rbf_affinity. Should accept (X1, X2=None, gamma=float) and return affinity matrix
        
    Returns:
        eigenvectors (torch.Tensor): shape (N, n_eig)
        eigenvalues (torch.Tensor): sorted in descending order, shape (n_eig,)
    
    Examples:
        >>> from ncut_pytorch import ncut_fn
        >>> import torch
        >>> features = torch.rand(10000, 100)
        >>> eigvec, eigval = ncut_fn(features, n_eig=20)
        >>> print(eigvec.shape, eigval.shape)  # (10000, 20) (20,)
    """
    config = NystromConfig()
    config.update(kwargs)

    # use GPU if available
    device = auto_device(X.device, device)

    # skip pytorch gradient computation if track_grad is False
    prev_grad_state = torch.is_grad_enabled()
    torch.set_grad_enabled(track_grad)

    # subsample for nystrom approximation
    n_sample = min(config.n_sample, int(X.shape[0]*config.n_sample_max_ratio))
    nystrom_indices = farthest_point_sampling(X, n_sample=n_sample, device=device)
    nystrom_X = X[nystrom_indices].to(device)

    # find optimal gamma for affinity matrix
    if gamma is None:
        gamma = find_gamma_by_degree_after_fps(nystrom_X, d_gamma, affinity_fn)

    # compute Ncut on the nystrom sampled subgraph
    A = affinity_fn(nystrom_X, gamma=gamma)
    nystrom_eigvec, eigval = _plain_ncut(A, n_eig)

    if no_propagation:
        torch.set_grad_enabled(prev_grad_state)
        return nystrom_eigvec, eigval, nystrom_indices, gamma

    # propagate eigenvectors from subgraph to full graph
    eigvec = nystrom_propagate(
        nystrom_eigvec,
        X,
        nystrom_X,
        n_neighbors=config.n_neighbors,
        n_sample=config.n_sample2,
        gamma=gamma,
        matmul_chunk_size=config.matmul_chunk_size,
        device=device,
        move_output_to_cpu=config.move_output_to_cpu,
        track_grad=track_grad,
        affinity_fn=affinity_fn,
    )

    # post-hoc orthogonalization
    if make_orthogonal:
        eigvec = gram_schmidt(eigvec)

    torch.set_grad_enabled(prev_grad_state)

    return eigvec, eigval


def _plain_ncut(
        A: torch.Tensor,
        n_eig: int = 100,
):
    # normalization; A = D^(-1/2) A D^(-1/2)
    A = normalize_affinity(A)

    eigvec, eigval, _ = svd_lowrank(A, n_eig)

    eigvec = correct_rotation(eigvec)

    return eigvec, eigval


def nystrom_propagate(
        nystrom_out: torch.Tensor,
        X: torch.Tensor,
        nystrom_X: torch.Tensor,
        gamma: float = 1.0,
        track_grad: bool = False,
        device: str = None,
        return_indices: bool = False,
        affinity_fn: Callable[[torch.Tensor, torch.Tensor, float], torch.Tensor] = rbf_affinity,
        **kwargs,
) -> Union[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
    """propagate output from nystrom sampled nodes to all nodes,
    use a weighted sum of the nearest neighbors to propagate the output.

    Args:
        nystrom_out (torch.Tensor): output from nystrom sampled nodes, shape (m, D)
        X (torch.Tensor): input features for all nodes, shape (N, D)
        nystrom_X (torch.Tensor): input features from nystrom sampled nodes, shape (m, D)
        gamma (float): affinity parameter, default 1.0
        track_grad (bool): keep track of pytorch gradients, default False
        device (str): device to use for computation, if 'auto', will detect GPU automatically
        affinity_fn (callable): affinity function, default rbf_affinity. Should accept (X1, X2=None, gamma=float) and return affinity matrix

    Returns:
        torch.Tensor: output propagated by nearest neighbors, shape (N, D)
    """

    config = NystromConfig()
    config.update(kwargs)

    # skip pytorch gradient computation if track_grad is False
    prev_grad_state = torch.is_grad_enabled()
    torch.set_grad_enabled(track_grad)

    device = auto_device(nystrom_out.device, device)
    indices = farthest_point_sampling(nystrom_out, config.n_sample2, device=device)
    nystrom_out = nystrom_out[indices].to(device)
    nystrom_X = nystrom_X[indices].to(device)
    
    D = affinity_fn(nystrom_X, gamma=gamma).mean(1)

    all_outs = []
    n_chunk = config.matmul_chunk_size
    n_neighbors = int(min(config.n_neighbors, len(indices)*config.n_neighbors_max_ratio))
    n_neighbors = max(n_neighbors, 4)
    for i in range(0, X.shape[0], n_chunk):
        end = min(i + n_chunk, X.shape[0])

        _Ai = affinity_fn(X[i:end].to(device), nystrom_X, gamma=gamma)
        _Ai, _indices = keep_topk_per_row(_Ai, n_neighbors)  # (n, n_neighbors)
        _Di = D[_indices].sum(1)
        _Ai = _Ai / _Di[:, None]

        weights = _Ai[..., None]  # (n, n_neighbors, 1)
        neighbors = nystrom_out[_indices.flatten()]
        neighbors = neighbors.reshape(-1, n_neighbors, nystrom_out.shape[-1])  # (n, n_neighbors, d)
        out = weights * neighbors  # (n, n_neighbors, d)
        out = out.sum(dim=1)  # (n, d)

        if config.move_output_to_cpu:
            out = out.to("cpu")
        all_outs.append(out)

    all_outs = torch.cat(all_outs, dim=0)

    torch.set_grad_enabled(prev_grad_state)

    if return_indices:
        return all_outs, indices
    
    return all_outs






# Credits: https://github.com/tldr-group/HR-Dv2

"""Patch the various methods of classes and subclasses in the Vision Transformer to add ]
new features like overlapping patches and attention visualisation"""

import math
from typing import Tuple, Callable, Literal

import torch
import torch.nn.functional as F

AttentionOptions = Literal["q", "k", "v", "o", "none"]


def get_qkvo_per_head(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    x_a: torch.Tensor,
    which: AttentionOptions,
    drop_fn: Callable,
) -> torch.Tensor:
    """Return either the mean q, k, v per head for tokens or the attn for the CLS
    token per head. Note that using mem eff attn means we need to explicitly
    recompute the attn tensor. This is expensive and scales poorly with image
    resolution.

    https://github.com/facebookresearch/dinov2/pull/306
    https://github.com/facebookresearch/dinov2/issues/90
    https://github.com/facebookresearch/xformers/issues/730#issuecomment-1518740489

    Nt=N_tokens (patches, [CLS] and [REGs]), Nh=N_heads and C is output embed dim of
    model, for vit_tiny C=384, Nh=6

    :param q: query tensor, shape [B, Nt, Nh, C //Nh]
    :type q: torch.Tensor
    :param k: key tensor, shape [B, Nt, Nh, C //Nh]
    :type k: torch.Tensor
    :param v: value tensor, shape [B, Nt, Nh, C //Nh]
    :type v: torch.Tensor
    :param x_a: output of mem eff attn tensor
    :type x_a: torch.Tensor
    :param which: choice of which of qkvo to extract
    :type which: AttentionOptions
    :raises Exception: if "none" attention option passed
    :return: desired primitive of shape [B, Nt, Nh]
    :rtype: torch.Tensor
    """
    prims, mapping = [q, k, v], "qkv"
    if which == "q" or which == "k" or which == "v":
        prim = prims[mapping.index(which)]
        per_head = prim.sum(dim=-1)
    elif which == "o":
        B, T, nH, _ = q.shape
        # we only care about (and compute) attn of [CLS] token
        x_a_cls = x_a[:, 0, :, :]
        x_a_cls = x_a_cls[:, None, :, :]
        cls_attn = x_a_cls.permute(0, 2, 1, 3) @ v.permute(0, 2, 3, 1)
        cls_attn = cls_attn.squeeze(2).permute(0, 2, 1)
        per_head = cls_attn.reshape([B, T, nH])
    else:
        raise Exception("not valid attention option")
    return per_head


class Patch:
    @staticmethod
    def _fix_pos_enc(patch_size: int, stride_hw: Tuple[int, int]) -> Callable:
        """Creates a method for position encoding interpolation, used to overwrite
        the original method in the DINO/DINOv2 vision transformer.
        Taken from https://github.com/ShirAmir/dino-vit-features/blob/main/extractor.py,
        added some bits from the Dv2 code in.

        :param patch_size: patch size of the model.
        :type patch_size: int
        :param stride_hw: A tuple containing the new height and width stride respectively.
        :type Tuple[int, int]
        :return: the interpolation method
        :rtype: Callable
        """

        def interpolate_pos_encoding(
            self, x: torch.Tensor, w: int, h: int
        ) -> torch.Tensor:
            previous_dtype = x.dtype
            npatch = x.shape[1] - 1
            N = self.pos_embed.shape[1] - 1
            if npatch == N and w == h:
                return self.pos_embed
            pos_embed = self.pos_embed.float()
            class_pos_embed = pos_embed[:, 0]
            patch_pos_embed = pos_embed[:, 1:]
            dim = x.shape[-1]
            # compute number of tokens taking stride into account
            w0: float = 1 + (w - patch_size) // stride_hw[1]
            h0: float = 1 + (h - patch_size) // stride_hw[0]
            assert (
                w0 * h0 == npatch
            ), f"""got wrong grid size for {h}x{w} with patch_size {patch_size} and
            #                               stride {stride_hw} got {h0}x{w0}={h0 * w0} expecting {npatch}"""
            # we add a small number to avoid floating point error in the interpolation
            # see discussion at https://github.com/facebookresearch/dino/issues/8
            w0, h0 = w0 + 0.1, h0 + 0.1
            patch_pos_embed = F.interpolate(
                patch_pos_embed.reshape(
                    1, int(math.sqrt(N)), int(math.sqrt(N)), dim
                ).permute(0, 3, 1, 2),
                scale_factor=(w0 / math.sqrt(N), h0 / math.sqrt(N)),
                mode="bicubic",
                align_corners=False,
                recompute_scale_factor=False,
            )
            assert (
                int(w0) == patch_pos_embed.shape[-2]
                and int(h0) == patch_pos_embed.shape[-1]
            )
            patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
            return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1).to(
                previous_dtype
            )

        return interpolate_pos_encoding

    """The next 5 methods are taken from user 'legel' on GitHub https://github.com/facebookresearch/dinov2/pull/306,
    who's written a pull request for Dv2 to output attention maps. The maintainers suggest using a hook, which is
    another option. Instead of editing the Dv2 source, I've used the same MethodType assignment procedure as above."""

    @staticmethod
    def _fix__attn() -> Callable:
        """Replaces normal 'forward()' method of the attention layer (block.attn) in the Dv2 model with
        an optional early return with attention.

        :return: the new forward method
        :rtype: Callable
        """

        def forward(self, x: torch.Tensor, return_attn: bool = False) -> torch.Tensor:
            B, N, C = x.shape
            qkv = (
                self.qkv(x)
                .reshape(B, N, 3, self.num_heads, C // self.num_heads)
                .permute(2, 0, 3, 1, 4)
            )

            q, k, v = qkv[0] * self.scale, qkv[1], qkv[2]
            attn = q @ k.transpose(-2, -1)

            attn = attn.softmax(dim=-1)
            attn = self.attn_drop(attn)

            if return_attn:
                return attn

            x = (attn @ v).transpose(1, 2).reshape(B, N, C)
            x = self.proj(x)
            x = self.proj_drop(x)

            return x

        return forward

    @staticmethod
    def _fix_attn_masking(attention_mask_ratio: float) -> Callable:
        """Replaces normal 'forward()' method of the memory efficient attention layer (block.attn)
        in the Dv2 model with an optional early return with attention. Used if xformers used.

        :return: the new forward method
        :rtype: Callable
        """

        def forward(
            self,
            x: torch.Tensor,
            attn_bias=None,
            attn_choice: AttentionOptions = "none",
        ) -> torch.Tensor:

            B, N, C = x.shape
            qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads)

            q, k, v = torch.unbind(qkv, 2) 

            # mask out some of the attention Keys (K) to save compute
            if attention_mask_ratio > 0 and attention_mask_ratio < 1:
                num_keys = int(N * attention_mask_ratio)
                mask_indices = torch.randperm(N-1)[:num_keys] + 1
                mask_indices = torch.cat((torch.tensor([0]), mask_indices)) # add the CLS token
                masked_k = k[:, mask_indices, :, :]
                masked_v = v[:, mask_indices, :, :]
                x = F.scaled_dot_product_attention(q.transpose(1, 2), 
                                                masked_k.transpose(1, 2), 
                                                masked_v.transpose(1, 2))
            else:
                x = F.scaled_dot_product_attention(q.transpose(1, 2), 
                                                k.transpose(1, 2), 
                                                v.transpose(1, 2), 
                                                attn_mask=attn_bias)
            x = x.transpose(1, 2)
            
            to_append: torch.Tensor
            if attn_choice != "none":
                to_append = get_qkvo_per_head(q, k, v, x, attn_choice, self.attn_drop)

            x = x.reshape([B, N, C])

            x = self.proj(x)
            x = self.proj_drop(x)

            if attn_choice != "none":
                x = torch.concat((x, to_append), dim=-1)
            return x

        return forward

    @staticmethod
    def _fix_block_forward_dino() -> Callable:
        """
        Replaces normal 'forward()' method of the block module to ensure the 'return_attn'
        flag is used in the attn forward layers.

        :return: the new forward method
        :rtype: Callable
        """

        def forward(
            self, x: torch.Tensor, attn_choice: AttentionOptions = "none"
        ) -> torch.Tensor:
            def attn_residual_func(x: torch.Tensor) -> torch.Tensor:
                # return self.ls1(self.attn(self.norm1(x)))
                return self.attn(self.norm1(x))

            def ffn_residual_func(x: torch.Tensor) -> torch.Tensor:
                # return self.ls2(self.mlp(self.norm2(x)))
                return self.mlp(self.norm2(x))

            if attn_choice != "none":
                nH = self.attn.num_heads
                ax = self.attn(self.norm1(x), attn_choice=attn_choice)
                xa, a = ax[:, :, :-nH], ax[:, :, -nH:]
                x = x + xa  # self.ls1(xa)
                x = x + ffn_residual_func(x)
                x = torch.concat((x, a), dim=-1)
                return x
            if self.training and self.sample_drop_ratio > 0.1:
                # the overhead is compensated only for a drop path rate larger than 0.1
                x = drop_add_residual_stochastic_depth(
                    x,
                    residual_func=attn_residual_func,
                    sample_drop_ratio=self.sample_drop_ratio,
                )
                x = drop_add_residual_stochastic_depth(
                    x,
                    residual_func=ffn_residual_func,
                    sample_drop_ratio=self.sample_drop_ratio,
                )
            elif self.training and self.sample_drop_ratio > 0.0:
                x = x + self.drop_path1(attn_residual_func(x))
                x = x + self.drop_path1(ffn_residual_func(x))  # FIXME: drop_path2
            else:
                x = x + attn_residual_func(x)
                x = x + ffn_residual_func(x)
            return x

        return forward

    @staticmethod
    def _fix_block_forward_dv2() -> Callable:
        """
        Replaces normal 'forward()' method of the block module to ensure the 'return_attn'
        flag is used in the attn forward layers.

        :return: the new forward method
        :rtype: Callable
        """

        def forward(
            self, x: torch.Tensor, attn_choice: AttentionOptions = "none"
        ) -> torch.Tensor:
            def attn_residual_func(x: torch.Tensor) -> torch.Tensor:
                return self.ls1(self.attn(self.norm1(x)))

            def ffn_residual_func(x: torch.Tensor) -> torch.Tensor:
                return self.ls2(self.mlp(self.norm2(x)))

            if attn_choice != "none":
                nH = self.attn.num_heads
                ax = self.attn(self.norm1(x), attn_choice=attn_choice)
                xa, a = ax[:, :, :-nH], ax[:, :, -nH:]
                x = x + self.ls1(xa)
                x = x + ffn_residual_func(x)
                x = torch.concat((x, a), dim=-1)
                return x
            if self.training and self.sample_drop_ratio > 0.1:
                # the overhead is compensated only for a drop path rate larger than 0.1
                x = drop_add_residual_stochastic_depth(
                    x,
                    residual_func=attn_residual_func,
                    sample_drop_ratio=self.sample_drop_ratio,
                )
                x = drop_add_residual_stochastic_depth(
                    x,
                    residual_func=ffn_residual_func,
                    sample_drop_ratio=self.sample_drop_ratio,
                )
            elif self.training and self.sample_drop_ratio > 0.0:
                x = x + self.drop_path1(attn_residual_func(x))
                x = x + self.drop_path1(ffn_residual_func(x))  # FIXME: drop_path2
            else:
                x = x + attn_residual_func(x)
                x = x + ffn_residual_func(x)
            return x

        return forward

    @staticmethod
    def _add_new_forward_features_dino() -> Callable:
        def forward_feats_attn(
            self, x, masks=None, attn_choice: AttentionOptions = "none"
        ):
            if isinstance(x, list):
                return self.forward_features_list(x, masks)

            x = self.prepare_tokens(x)  # prepare_tokens_with_masks(x, masks) for dv2

            for i, blk in enumerate(self.blocks):
                if i < len(self.blocks) - 1:
                    x = blk(x)
                else:
                    x = blk(x, attn_choice=attn_choice)

            if attn_choice != "none":
                x_feats = x[:, :, : -self.num_heads]
                # in our new function, the attn options are the last 6 channels of the features
                x_attn = x[:, :, -self.num_heads :]
            else:
                x_feats = x

            x_norm = self.norm(x_feats)
            out_dict = {
                "x_norm_clstoken": x_norm[:, 0],
                "x_norm_regtokens": x_norm[:, 1 : self.num_register_tokens + 1],
                "x_norm_patchtokens": x_norm[:, self.num_register_tokens + 1 :],
                "x_prenorm": x,
                "masks": masks,
            }

            if attn_choice != "none":
                out_dict["x_patchattn"] = x_attn[:, self.num_register_tokens + 1 :]

            return out_dict

        return forward_feats_attn

    @staticmethod
    def _add_new_forward_features_dv2() -> Callable:
        def forward_feats_attn(
            self, x, masks=None, attn_choice: AttentionOptions = "none"
        ):
            if isinstance(x, list):
                return self.forward_features_list(x, masks)

            x = self.prepare_tokens_with_masks(x, masks)

            for i, blk in enumerate(self.blocks):
                if i < len(self.blocks) - 1:
                    x = blk(x)
                else:
                    x = blk(x, attn_choice=attn_choice)

            if attn_choice != "none":
                x_feats = x[:, :, : -self.num_heads]
                # in our new function, the attn options are the last 6 channels of the features
                x_attn = x[:, :, -self.num_heads :]
            else:
                x_feats = x

            x_norm = self.norm(x_feats)
            out_dict = {
                "x_norm_clstoken": x_norm[:, 0],
                "x_norm_regtokens": x_norm[:, 1 : self.num_register_tokens + 1],
                "x_norm_patchtokens": x_norm[:, self.num_register_tokens + 1 :],
                "x_prenorm": x,
                "masks": masks,
            }

            if attn_choice != "none":
                out_dict["x_patchattn"] = x_attn[:, self.num_register_tokens + 1 :]

            return out_dict

        return forward_feats_attn


# here to avoid syntax errors
def drop_add_residual_stochastic_depth(
    x: torch.Tensor,
    residual_func: Callable[[torch.Tensor], torch.Tensor],
    sample_drop_ratio: float = 0.0,
) -> torch.Tensor:
    # 1) extract subset using permutation
    b, n, d = x.shape
    sample_subset_size = max(int(b * (1 - sample_drop_ratio)), 1)
    brange = (torch.randperm(b, device=x.device))[:sample_subset_size]
    x_subset = x[brange]

    # 2) apply residual_func to get residual
    residual = residual_func(x_subset)

    x_flat = x.flatten(1)
    residual = residual.flatten(1)

    residual_scale_factor = b / sample_subset_size

    # 3) add the residual
    x_plus_residual = torch.index_add(
        x_flat, 0, brange, residual.to(dtype=x.dtype), alpha=residual_scale_factor
    )
    return x_plus_residual.view_as(x)



from typing import Tuple

import torch
from torchvision import transforms
from .hires_dino import hires_dino
from .hires_dino import HighResDINO
from .transform import get_input_transform


def hires_dino_256(dtype: torch.dtype = torch.float16) \
        -> Tuple[HighResDINO, transforms.Compose]:
    model = hires_dino(dino_name="dino_vitb8",
                       stride=6,
                       shift_dists=[1, 2, 3],
                       flip_transforms=True,
                       chunk_size=6,
                       dtype=dtype,
                       feature_resolution=256)
    transform = get_input_transform(resize=256)
    return model, transform


def hires_dino_512(dtype: torch.dtype = torch.float16) \
        -> Tuple[HighResDINO, transforms.Compose]:
    model = hires_dino(dino_name="dino_vitb8",
                       stride=6,
                       shift_dists=[1, 2, 3],
                       flip_transforms=True,
                       chunk_size=4,
                       dtype=dtype,
                       feature_resolution=512)
    transform = get_input_transform(resize=512)
    return model, transform


def hires_dino_1024(dtype: torch.dtype = torch.float16) \
        -> Tuple[HighResDINO, transforms.Compose]:
    model = hires_dino(dino_name="dino_vitb8",
                       stride=6,
                       shift_dists=[1, 2, 3],
                       flip_transforms=True,
                       chunk_size=1,
                       dtype=dtype,
                       feature_resolution=1024)
    transform = get_input_transform(resize=1024)
    return model, transform


def hires_dinov2(dtype: torch.dtype = torch.float16) \
        -> Tuple[HighResDINO, transforms.Compose]:
    model = hires_dino(dino_name="dinov2_vitb14_reg",
                       stride=6,
                       shift_dists=[1, 2, 3],
                       flip_transforms=True,
                       chunk_size=1,
                       dtype=dtype,
                       feature_resolution=1008)
    transform = get_input_transform(resize=1008)
    return model, transform



URLS = {
    "dinov3_vits16": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vits16_pretrain_lvd1689m-08c60483.pth",
    "dinov3_vits16plus": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vits16plus_pretrain_lvd1689m-4057cbaa.pth",
    "dinov3_vitb16": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vitb16_pretrain_lvd1689m-73cec8be.pth",
    "dinov3_vitl16": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vitl16_pretrain_lvd1689m-8aa4cbdd.pth",
    "dinov3_vith16plus": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vith16plus_pretrain_lvd1689m-7c1da9a5.pth",
    "dinov3_vitl16_sat493m": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vitl16_pretrain_sat493m-eadcf0ff.pth",
    "dinov3_vitl16_dinotxt": "https://huggingface.co/huzey/mydv3/resolve/master/dinov3_vitl16_dinotxt_vision_head_and_text_encoder-a442d8f5.pth",
}

import torch
from torch import nn
import os
import subprocess


class Dinov3Backbone(nn.Module):
    def __init__(self, config="dinov3_vitl16"):
        super().__init__()
        # ckpt_path = download_by_wget(URLS[config])
        # dinov3 = torch.hub.load("facebookresearch/dinov3", config, weights=ckpt_path)
        if config == "dinov3_vitl16_sat493m":
            c = "dinov3_vitl16"
        else:
            c = config
        # if config == "dinov3_vitl16_dinotxt":
        #     config = "dinov3_vitl16"
        dinov3 = torch.hub.load("facebookresearch/dinov3", c, weights=URLS[config])
        self.model = dinov3

    def forward(self, x: torch.Tensor):
        return self.model.get_intermediate_layers(x, reshape=True)[0]  # b, c, h, w


def download_by_wget(url):
    # Get the torch hub directory
    torch_hub_dir = torch.hub._get_torch_home()
    cache_dir = os.path.join(torch_hub_dir, 'checkpoints')
    os.makedirs(cache_dir, exist_ok=True)

    model_filename = url.split('/')[-1].split('?')[0]
    output_file = os.path.join(cache_dir, model_filename)

    # Download the file using wget (with quotes around the URL to handle special characters)
    # print(f"Downloading model to {output_file}...")
    subprocess.run(['wget', '-O', output_file, url], check=True)
    return output_file


# Credits: https://github.com/tldr-group/HR-Dv2

# ==================== IMPORTS ====================
import torch
import torch.nn as nn
import numpy as np

from .patch import Patch

from types import MethodType
from typing import Literal, Union

AttentionOptions = Literal["q", "k", "v", "o", "none"]
DINONameOptions = Literal["dino_vitb8", "dino_vits8", "dino_vitb16", "dinov2_vitb14", "dinov2_vitb14_reg"]

# ==================== MODULE ====================


class LowResDINO(nn.Module):
    def __init__(
        self,
        dino_name: DINONameOptions = "dino_vitb8",
        dtype: Union[torch.dtype, int] = torch.float16,
        track_grad: bool = False,
        attention_mask_ratio: float = 0.1,
    ) -> None:
        super().__init__()

        self.dinov2: nn.Module
        if "dinov2" in dino_name:
            hub_path = "facebookresearch/dinov2"
            self.dinov2 = torch.hub.load(hub_path, dino_name, verbose=False)
        elif "dino" in dino_name:
            hub_path = "facebookresearch/dino:main"
            self.dinov2 = torch.hub.load(hub_path, dino_name, verbose=False)
        else:
            raise ValueError(f"Invalid model name: {dino_name}")

        self.dinov2.eval()
        
        if "dinov2" not in dino_name:
            self.dinov2.num_heads = 6  # type: ignore
            self.dinov2.num_register_tokens = 0  # type: ignore

        # If we want to save memory, change to float16
        if type(dtype) == int:
            dtype = torch.float16 if dtype == 16 else torch.float32

        self.dtype = dtype
        if dtype != torch.float32:
            self = self.to(dtype)
        self.track_grad = track_grad  # off by default to save memory

        self.attention_mask_ratio = attention_mask_ratio
        self.patch_last_block(self.dinov2, dino_name)

    def patch_last_block(self, dino_model: nn.Module, dino_name: str) -> None:
        """Patch the final block of the dino model to add attention return code.

        :param dino_model: DINO or DINOv2 model
        :type dino_model: nn.Module
        """
        final_block = dino_model.blocks[-1]  # type: ignore
        attn_block = final_block.attn  # type: ignore
        # hilariously this also works for dino i.e we can patch dino's attn block forward to
        # use the memeory efficienty attn like in dinov2
        attn_block.forward = MethodType(Patch._fix_attn_masking(self.attention_mask_ratio), attn_block)
        if "dinov2" in dino_name:
            final_block.forward = MethodType(Patch._fix_block_forward_dv2(), final_block)  # type: ignore
            dino_model.forward_feats_attn = MethodType(  # type: ignore
                Patch._add_new_forward_features_dv2(), dino_model
            )
        elif "dino" in dino_name:
            for i, blk in enumerate(dino_model.blocks):
                blk.forward = MethodType(Patch._fix_block_forward_dino(), blk)
                attn_block = blk.attn
                attn_block.forward = MethodType(Patch._fix_attn_masking(self.attention_mask_ratio), attn_block)
            final_block.forward = MethodType(Patch._fix_block_forward_dino(), final_block)  # type: ignore
            dino_model.forward_feats_attn = MethodType(  # type: ignore
                Patch._add_new_forward_features_dino(), dino_model
            )
        else:
            for i, blk in enumerate(dino_model.blocks):
                blk.forward = MethodType(Patch._fix_block_forward_dino(), blk)
                attn_block = blk.attn
                attn_block.forward = MethodType(Patch._fix_attn_masking(self.attention_mask_ratio), attn_block)
            final_block.forward = MethodType(Patch._fix_block_forward_dino(), final_block)  # type: ignore
            dino_model.forward_feats_attn = MethodType(  # type: ignore
                Patch._add_new_forward_features_vit(), dino_model
            )

    def _forward_batch(
        self,
        x: torch.Tensor,
        attn_choice: AttentionOptions = "none",
    ) -> torch.Tensor:
        """Feed input img $x through network and get low-res features.

        :param x: batched image tensor, (b, c, h, w)
        :param attn_choice: choice of attention, "none" or "q", "k", "v", "o"
        :return: low-res features, (b, c, h//k, w//k)
        """
        if self.dtype != torch.float32:  # cast (i.e to f16)
            x = x.type(self.dtype)

        out_dict = self.dinov2.forward_feats_attn(x, None, attn_choice)  # type: ignore
        if attn_choice != "none":
            feats, attn = out_dict["x_norm_patchtokens"], out_dict["x_patchattn"]
            features_batch = torch.concat((feats, attn), dim=-1)
        else:
            features_batch = out_dict["x_norm_patchtokens"]

        if self.dtype != torch.float32:  # cast (i.e to f16)
            features_batch = features_batch.type(self.dtype)
            
        b, l, c = features_batch.shape
        hw = np.sqrt(l).astype(int)
        features_batch = features_batch.reshape(b, hw, hw, c)
        features_batch = features_batch.permute(0, 3, 1, 2)
        return features_batch            
    
    def forward(
        self,
        x: torch.Tensor,
        attn_choice: AttentionOptions = "none",
        move_to_cpu: bool = True,
    ) -> torch.Tensor:
        """Feed input img $x through network and get low-res features.

        :param x: batched image tensor, (b, c, h, w)
        :param attn_choice: choice of attention, "none" or "q", "k", "v", "o"
        :return: low-res features, (b, c, h//k, w//k)
        :rtype: torch.Tensor
        """
        if self.track_grad:
            out = self._forward_batch(x, attn_choice)
        else:
            with torch.no_grad():
                out = self._forward_batch(x, attn_choice)
                if move_to_cpu:
                    out = out.cpu()
        return out

    


